@@ -54,12 +54,19 @@ except ImportError:
 	print "%s needs gnome-python 2.10!" % NAME
 	sys.exit(1)
 
+# only so I can test SC in my older boxes
+if 'filename_display_name' not in dir(gobject):
+	def __fake_display_name(name):
+		return name
+	gobject.filename_display_name = __fake_display_name
+
 # GStreamer
 try:
 	# 0.10
 	import pygst
 	pygst.require('0.10')
 	import gst
+	
 except ImportError:
 	print "%s needs python-gstreamer 0.10!" % NAME
 	sys.exit(1)
@@ -121,7 +128,7 @@ mime_whitelist = (
 # custom filename patterns
 english_patterns = "Artist Album Title Track Total Genre Date Year"
 
-# traductors: only if it does make sense.
+# traductors: These are the custom filename patterns. Only if it does make sense.
 locale_patterns = _("Artist Album Title Track Total Genre Date Year")
 
 patterns_formats = (
@@ -186,20 +193,24 @@ def vfs_walk(uri):
 	try:
 		dirlist = gnomevfs.open_directory(uri)
 	except:
-		pass
 		log(_("skipping: '%s'") % uri)
 		return filelist
 		
 	for file_info in dirlist:
-		if file_info.name[0] == ".":
-			continue
+		try:
+			if file_info.name[0] == ".":
+				continue
 
-		if file_info.type == gnomevfs.FILE_TYPE_DIRECTORY:
-			filelist.extend(
-				vfs_walk(uri.append_path(file_info.name)) )
+			if file_info.type == gnomevfs.FILE_TYPE_DIRECTORY:
+				filelist.extend(
+					vfs_walk(uri.append_path(file_info.name)) )
 
-		if file_info.type == gnomevfs.FILE_TYPE_REGULAR:
-			filelist.append( str(uri.append_file_name(file_info.name)) )
+			if file_info.type == gnomevfs.FILE_TYPE_REGULAR:
+				filelist.append( str(uri.append_file_name(file_info.name)) )
+		except ValueError:
+			# this can happen when you do not have sufficent
+			# permissions to read file info.
+			log(_("skipping: '%s'") % uri)
 	return filelist
 
 def vfs_makedirs(path_to_create):
@@ -288,7 +299,6 @@ if gst.element_factory_find("gnomevfssrc"):
 	gstreamer_sink = "gnomevfssink"
 	encode_filename = vfs_encode_filename
 	use_gnomevfs = True
-	print "  using gnomevfssrc"
 else:
 	gstreamer_source = "filesrc"
 	gstreamer_sink = "filesink"
@@ -670,59 +680,80 @@ class TaskQueue(BackgroundTask):
 	def __init__(self):
 		BackgroundTask.__init__(self)
 		self.tasks = []
-		self.running = False
-		self.tasks_current = 0
+		self.all_tasks = []
+		self.running = None
+		self.tasks_done = 0
 		self.tasks_number = 0
 
 	def is_running(self):
-		return self.running
+		if self.running:
+			return True
+		return False
 
 	def add(self, task):
+		#print 'adding task:', task
 		self.tasks.append(task)
+		self.all_tasks.append(task)
 		self.tasks_number += 1
 		
 	def get_current_task(self):
-		if self.tasks:
-			return self.tasks[0]
+		if self.running:
+			return self.running[0]
 		else:
 			return None
 
+	def start_next_task(self):
+		#print 'start next tasks:'
+		#self.running = []
+		to_start = get_option('jobs') - len(self.running)
+		#print 'trying to start:', to_start
+		for i in range(to_start):
+			try:
+				task = self.tasks.pop()
+			except IndexError:
+				return
+			self.running.append(task)
+			task.setup()
+
 	def setup(self):
 		""" BackgroundTask setup callback """
-		self.running = True
+		self.running = []
 		self.start_time = time.time()
+		self.tasks_done = 0
 
-		if self.tasks:
-			self.tasks[0].setup()
-			self.setup_hook(self.tasks[0])
+		self.start_next_task()
+		if self.running:
+			[self.setup_hook(task) for task in self.running]
 
 			
 	def work(self):
 		""" BackgroundTask work callback """
-		if self.tasks:
-			ret = self.tasks[0].work()
-			self.work_hook(self.tasks[0])
-			if not ret:
-				self.finish_hook(self.tasks[0])
-				self.tasks[0].finish()
-				self.tasks = self.tasks[1:]
-				if self.tasks:
-					self.tasks_current += 1
-					self.tasks[0].setup()
-		return len(self.tasks) > 0
+		if self.running:
+			self.work_hook(self.running)
+			for task in self.running:
+				ret = task.work()
+				if not ret:
+					self.tasks_done += 1
+					self.finish_hook(task)
+					task.finish()
+					self.running.remove(task)
+					self.start_next_task()
+			return True
+		return False
 
 	def finish(self):
 		""" BackgroundTask finish callback """
-		self.running = False
+		self.running = None 
 		log("Queue done in %ds" % (time.time() - self.start_time))
 		self.queue_ended()
 		self.tasks_number = 0
 
+
 	def stop(self):
-		if self.tasks:
-			self.tasks[0].stop()
+		if self.running:
+			[task.stop() for task in self.running]
 		BackgroundTask.stop(self)
-		self.running = False
+		self.running = None
 		self.tasks = []
 		self.tasks_number = 0
 
@@ -776,13 +807,7 @@ class Pipeline(BackgroundTask):
 		self.play()
 	
 	def work(self):
-		#if self.pipeline.get_state() == gst.STATE_NULL:
-		#  log("error: pipeline.state == null")
-			#return False
-		#print "work:", self
-		#time.sleep(0.01)
 		if self.eos:
-			#print "	got eos:", self.sound_file.get_filename_for_display()
 			return False
 		return True
 
@@ -799,6 +824,10 @@ class Pipeline(BackgroundTask):
 		self.signals.append( (name, signal, callback,) )
 
 	def toggle_pause(self, paused):
+		if not self.pipeline:
+			debug("toggle_pause(): pipeline is None !")
+			return
+
 		if paused:
 			self.pipeline.set_state(gst.STATE_PAUSED)
 		else:
@@ -806,15 +835,44 @@ class Pipeline(BackgroundTask):
 
 	def found_tag(self, decoder, something, taglist):
 		pass
+		
+	def install_plugin_cb(self, result):
+		if result == gst.pbutils.INSTALL_PLUGINS_SUCCESS:
+			gst.update_registry()
+			self.parsed = False
+			self.play()
+			return
+		self.finish()
+		if result == gst.pbutils.INSTALL_PLUGINS_USER_ABORT:
+			dialog = gtk.MessageDialog(parent=None, flags=gtk.DIALOG_MODAL, 
+				type=gtk.MESSAGE_INFO, 
+				buttons=gtk.BUTTONS_OK, 
+				message_format="Plugin installation aborted.")
+			dialog.run()
+			dialog.hide()
+			return
 
+		error.show("Error", "failed to install plugins: %s" % markup_escape(str(result)))
+		
 	def on_message(self, bus, message):
 		t = message.type
+		import gst
 		if t == gst.MESSAGE_ERROR:
 			err, debug = message.parse_error()
 			self.eos = True
 			self.error = err
-			#log("error: %s (%s)" % (err,
-			#	self.sound_file.get_filename_for_display()))
+			log("error: %s (%s)" % (err,
+				self.sound_file.get_filename_for_display()))
+				
+		elif t == gst.MESSAGE_ELEMENT:
+			st = message.structure
+			if st and st.get_name().startswith('missing-'):
+				self.pipeline.set_state(gst.STATE_NULL)
+				if gst.pygst_version >= (0, 10, 10):
+					import gst.pbutils
+					detail = gst.pbutils.missing_plugin_message_get_installer_detail(message)
+					ctx = gst.pbutils.InstallPluginsContext()
+					gst.pbutils.install_plugins_async([detail], ctx, self.install_plugin_cb)
 			#error.show("GStreamer Error", "%s\nfile: '%s'" % (err, 
 			#	self.sound_file.get_filename_for_display()))
 		elif t == gst.MESSAGE_EOS:
@@ -826,12 +884,20 @@ class Pipeline(BackgroundTask):
 	def play(self):
 		if not self.parsed:
 			debug("launching: '%s'" % self.command)
-			self.pipeline = gst.parse_launch(self.command)
-			for name, signal, callback in self.signals:
-				self.pipeline.get_by_name(name).connect(signal,callback)
-			self.parsed = True
-	
-		bus = self.pipeline.get_bus()
+			try:
+				self.pipeline = gst.parse_launch(self.command)
+				bus = self.pipeline.get_bus()
+				for name, signal, callback in self.signals:
+					if name:
+						self.pipeline.get_by_name(name).connect(signal,callback)
+					else:
+						bus.connect(signal,callback)
+				self.parsed = True
+			except gobject.GError, e:
+				error.show("GStreamer error when creating pipeline", str(e))
+				self.eos = True # TODO
+				return
+
 		bus.add_signal_watch()
 		watch_id = bus.connect('message', self.on_message)
 		self.watch_id = watch_id
@@ -840,7 +906,7 @@ class Pipeline(BackgroundTask):
 
 	def stop_pipeline(self):
 		if not self.pipeline:
-			log("pipeline already stopped!")
+			debug("pipeline already stopped!")
 			return
 		bus = self.pipeline.get_bus()
 		bus.disconnect(self.watch_id)
@@ -919,13 +985,15 @@ class Decoder(Pipeline):
 
 	def _buffer_probe(self, pad, buffer):
 		"""buffer probe callback used to get real time since the beginning of the stream"""
-		if time.time() > self.time + 0.1:
-			self.time = time.time()
-			self.position = float(buffer.timestamp) / gst.SECOND
-
 		if buffer.timestamp == gst.CLOCK_TIME_NONE:
-			debug("removing probe")
+			debug("removing buffer probe")
 			pad.remove_buffer_probe(self.probe_id)
+			return False
+
+		#if time.time() > self.time + 0.1:
+		#	self.time = time.time()
+		self.position = float(buffer.timestamp) / gst.SECOND
+
 		return True
 	
 	def new_decoded_pad(self, decoder, pad, is_last):
@@ -933,8 +1001,6 @@ class Decoder(Pipeline):
 		self.probe_id = pad.add_buffer_probe(self._buffer_probe)
 		self.processing = True
 		self.query_duration()
-		#self.sound_file.duration = self.pipeline.query_duration(gst.FORMAT_TIME)[0] / gst.SECOND
-		#print "new_decoded_pad duration:", self.sound_file.duration
 
 	def get_sound_file(self):
 		return self.sound_file
@@ -963,10 +1029,16 @@ class TagReader(Decoder):
 		self.found_tags = False
 		self.run_start_time = 0 
 		self.add_command("fakesink")
+		self.add_signal(None, "message::state-changed", self.on_state_changed)
 
 	def set_found_tag_hook(self, found_tag_hook):
 		self.found_tag_hook = found_tag_hook
 
+	def on_state_changed(self, bus, message):
+		prev, new, pending = message.parse_state_changed()
+		if new == gst.STATE_PLAYING:
+			debug("TagReading done...")
+			self.finish()
 
 	def found_tag(self, decoder, something, taglist):
 		#debug("found_tags:", self.sound_file.get_filename_for_display())
@@ -989,18 +1061,25 @@ class TagReader(Decoder):
 			pass
 
 	def work(self):
+		if not self.pipeline:
+			return False
+		
 		if not self.run_start_time:
 			if self.sound_file.mime_type in tag_blacklist:
 				log("%s: type is %s, tag reading blacklisted" % (self.sound_file.get_filename_for_display(), self.sound_file.mime_type))
 				return False
 			self.run_start_time = time.time()
+			
+		if self.pipeline.get_state() != gst.STATE_PLAYING:
+			self.run_start_time = time.time()
+
 		if time.time()-self.run_start_time > 5:
 			# stop looking for tags after 5s 
 			return False
 		return Decoder.work(self) and not self.found_tags
 
 	def finish(self):
-		Decoder.finish(self)
+		Pipeline.finish(self)
 		self.sound_file.tags_read = True
 		if self.found_tag_hook:
 			gobject.idle_add(self.found_tag_hook, self)
@@ -1195,19 +1274,8 @@ class FileList:
 
 		if mime_id >= 0 and mime_id < len(self.drop_mime_types):
 			file_list = []
-			for uri in selection.data.split("\n"):
-				uri = uri.strip()
-				if uri:
-					info = gnomevfs.get_file_info(uri, gnomevfs.FILE_INFO_DEFAULT)
-					if info.type == gnomevfs.FILE_TYPE_DIRECTORY:
-						file_list.extend(vfs_walk(gnomevfs.URI(uri)))
-					else:
-						file_list.append(uri)
+			self.add_uris([uri.strip() for uri in selection.data.split("\n")])
 			context.finish(True, False, time)
-			base = os.path.commonprefix(file_list)
-			#[self.add_file(SoundFile(base, uri[len(base):])) for uri in file_list]
-			#[self.add_file(SoundFile(uri, base)) for uri in file_list]
-			self.add_uris(file_list)
 
 	def get_files(self):
 		files = []
@@ -1222,7 +1290,7 @@ class FileList:
 		return files
 	
 	def found_type(self, sound_file, mime):
-		#debug("found_type", sound_file.get_filename())
+		debug("found_type", sound_file.get_filename())
 
 		self.append_file(sound_file)
 		self.window.set_sensitive()
@@ -1237,8 +1305,12 @@ class FileList:
 	def add_uris(self, uris, base=None, filter=None):
 
 		files = []
-
+		
 		for uri in uris:
+			if uri.startswith('cdda:'):
+				error.show("Cannot read from Audio CD.",
+					"Use SoundJuicer Audio CD Extractor instead.")
+				return
 			try:
 				info = gnomevfs.get_file_info(gnomevfs.URI(uri))
 			except gnomevfs.NotFoundError:
@@ -1247,6 +1319,15 @@ class FileList:
 			except gnomevfs.InvalidURIError:
 				log('unvalid uri: \'%s\'' % uri)
 				continue
+			except gnomevfs.AccessDeniedError:
+				log('access denied: \'%s\'' % uri)
+				continue
+			except TypeError, e:
+				log('error: %s (%s)' % (e, uri))
+				continue
+			except :
+				log('error in get_file_info: %s' % (uri))
+				continue
 
 			if info.type == gnomevfs.FILE_TYPE_DIRECTORY:
 				filelist = vfs_walk(gnomevfs.URI(uri))
@@ -1268,11 +1349,12 @@ class FileList:
 				continue 
 			self.filelist[sound_file.get_uri()] = True
 
+			#self.found_type(sound_file, "test")
 			typefinder = TypeFinder(sound_file)
 			typefinder.set_found_type_hook(self.found_type)
 			self.typefinders.add(typefinder)
 
-		if not self.typefinders.is_running():
+		if files and not self.typefinders.is_running():
 			self.typefinders.queue_ended = self.typefinder_queue_ended
 			self.typefinders.run()
 
@@ -1324,6 +1406,7 @@ class FileList:
 		fields["META"] = sound_file
 		fields["filename"] = sound_file.get_filename_for_display()
 
+		# TODO: SLOW!
 		for i in self.model:
 			if i[1] == sound_file:
 				i[0] = self.format_cell(sound_file)
@@ -1865,7 +1948,7 @@ class ConverterQueue(TaskQueue):
 			gnomevfs.get_file_info(gnomevfs.URI((output_filename)))
 		except gnomevfs.NotFoundError:
 			exists = False
-		except gnomevfs.InvalidURIError:
+		except :
 			log("Invalid URI: '%s'" % output_filename)
 			return
 		
@@ -1934,33 +2017,36 @@ class ConverterQueue(TaskQueue):
 		c.got_duration = False
 		#self.total_duration += c.get_duration()
 
-	def work_hook(self, task):
-		gobject.idle_add(self.set_progress, (task))
+	def work_hook(self, tasks):
+		gobject.idle_add(self.set_progress, (tasks))
 
 	def get_progress(self, task):
 		return (self.duration_processed + task.get_position()) / self.total_duration
 
-	def set_progress(self, task):
-		t = self.get_current_task()
-		t = task 
-		f = ""
-		if t:
-			f = t.sound_file.get_filename_for_display()
+	def set_progress(self, tasks):
+		filename = ""
+		if tasks and tasks[0]:
+			filename = tasks[0].sound_file.get_filename_for_display()
 
-		for t in self.tasks:
-			if not t.got_duration:
-				duration = t.sound_file.duration
-				#duration = t.get_duration()
+		# try to get all tasks durations
+		total_duration = self.total_duration
+		for task in self.all_tasks:
+			if not task.got_duration:
+				duration = task.sound_file.duration
 				if duration: 
 					self.total_duration += duration
-					t.got_duration = True
+					task.got_duration = True
+				else:
+					total_duration = 0 
 
-		if task.converting :
-			position = task.get_position()
-		else:
-			position = 0
+		position = 0
+		for task in tasks:
+			if task.converting :
+				position += task.get_position()
+
+		#print self.duration_processed, position, total_duration
 		self.window.set_progress(self.duration_processed + position,
-							 self.total_duration, f)
+							 total_duration, filename)
 		return False
 
 	def finish_hook(self, task):
@@ -2337,7 +2423,7 @@ class SoundConverterWindow:
 		self._lock_convert_button = False
 	
 	def display_progress(self, remaining):
-		self.progressbar.set_text(_("Converting file %d of %d  (%s)") % ( self.converter.tasks_current+1, self.converter.tasks_number, remaining ))
+		self.progressbar.set_text(_("Converting file %d of %d  (%s)") % ( self.converter.tasks_done+1, self.converter.tasks_number, remaining ))
 	
 	def set_progress(self, done_so_far, total, current_file=""):
 		if (total==0) or (done_so_far==0):
@@ -2467,6 +2553,31 @@ def cli_convert_main(input_files):
 	if not get_option("quiet"):
 		progress.clear()
 
+def cpuCount():
+	'''
+	Returns the number of CPUs in the system.
+	(from pyprocessing)
+	'''
+	if sys.platform == 'win32':
+		try:
+			num = int(os.environ['NUMBER_OF_PROCESSORS'])
+		except (ValueError, KeyError):
+			num = 0
+	elif sys.platform == 'darwin':
+		try:
+			num = int(os.popen('sysctl -n hw.ncpu').read())
+		except ValueError:
+			num = 0
+	else:
+		try:
+			num = os.sysconf('SC_NPROCESSORS_ONLN')
+		except (ValueError, OSError, AttributeError):
+			num = 0
+	if num >= 1:
+		return num
+	else:
+		return 1
+		#raise NotImplementedError, 'cannot determine number of cpus'
 
 settings = {
 	"mode": "gui",
@@ -2474,6 +2585,7 @@ settings = {
 	"debug": False,
 	"cli-output-type": "audio/x-vorbis",
 	"cli-output-suffix": ".ogg",
+	"jobs": cpuCount(),
 }
 
 
@@ -2515,7 +2627,7 @@ options = [
 	 _("Be quiet. Don't write normal output, only errors.")),
 
 	("d", "debug", lambda optarg: set_option("debug", True),
-	 _("Print additionnal debug information")),
+	 _("Print additional debug information")),
 
 	("s:", "suffix=", lambda optarg: set_option("cli-output-suffix", optarg),
 	 _("Set the output filename suffix for batch mode. The default is \n %s . Note that the suffix does not affect\n the output MIME type.") % get_option("cli-output-suffix")),
@@ -2523,6 +2635,9 @@ options = [
 	("t", "tags", lambda optarg: set_option("mode", "tags"),
 	 _("Show tags for input files instead of converting them. This indicates \n command line batch mode and disables the graphical user interface.")),
 
+	("j", "jobs=", lambda optarg: set_option("jobs", optarg),
+	 _("Force number of concurrent conversions.")),
+
 	]
 
 
@@ -2530,7 +2645,12 @@ def main():
 	shortopts = "".join(map(lambda opt: opt[0], options))
 	longopts = map(lambda opt: opt[1], options)
 	
-	opts, args = getopt.getopt(sys.argv[1:], shortopts, longopts)
+	try:
+		opts, args = getopt.getopt(sys.argv[1:], shortopts, longopts)
+	except getopt.GetoptError, error:
+		print 'Error: ', error
+		sys.exit(1)
+
 	for opt, optarg in opts:
 		for tuple in options:
 			short = "-" + tuple[0][:1]
@@ -2540,13 +2660,12 @@ def main():
 			if opt in [short, long]:
 				tuple[2](optarg)
 				break
-	if 0:
-		print
-		for key in settings:
-			print key, settings[key]
-		return
 
 	args = map(filename_to_uri, args)
+
+	jobs = int(get_option('jobs'))
+	set_option('jobs', jobs)
+	print '  using %d thread(s)' % get_option('jobs')
 	
 	if get_option("mode") == "gui":
 		gui_main(args)
